@@ -22,22 +22,21 @@
 */
 
 #include <QCoreApplication>
+#include <QDebug>
 #include <QFile>
 #include <QFileInfo>
-#include <QTextStream>
-#include <QVariant>
-#include <QXmlStreamReader>
 #include <QJsonDocument>
 #include <QRegularExpression>
-#include <QDebug>
+#include <QVariant>
+#include <QXmlStreamReader>
 
 #ifdef QT_XMLPATTERNS_LIB
 #include <QXmlSchema>
 #include <QXmlSchemaValidator>
 #endif
 
-namespace {
-
+namespace
+{
 QStringList readListing(const QString &fileName)
 {
     QFile file(fileName);
@@ -57,8 +56,8 @@ QStringList readListing(const QString &fileName)
     }
 
     if (xml.hasError()) {
-        qWarning() << "XML error while reading" << fileName << " - "
-            << qPrintable(xml.errorString()) << "@ offset" << xml.characterOffset();
+        qWarning() << "XML error while reading" << fileName << " - " << qPrintable(xml.errorString()) << "@ offset" << xml.characterOffset();
+        listing.clear();
     }
 
     return listing;
@@ -69,7 +68,7 @@ QStringList readListing(const QString &fileName)
  * @param extensions extensions string to check
  * @return valid?
  */
-bool checkExtensions(QString extensions)
+bool checkExtensions(const QString &extensions)
 {
     // get list of extensions
     const QStringList extensionParts = extensions.split(QLatin1Char(';'), QString::SkipEmptyParts);
@@ -80,7 +79,7 @@ bool checkExtensions(QString extensions)
     }
 
     // check that only valid wildcard things are inside the parts
-    for (const auto& extension : extensionParts) {
+    for (const auto &extension : extensionParts) {
         for (const auto c : extension) {
             // eat normal things
             if (c.isDigit() || c.isLetter()) {
@@ -114,10 +113,10 @@ bool checkRegularExpression(const QString &hlFilename, QXmlStreamReader &xml)
 {
     if (xml.name() == QLatin1String("RegExpr") || xml.name() == QLatin1String("emptyLine")) {
         // get right attribute
-        const QString string (xml.attributes().value((xml.name() == QLatin1String("RegExpr")) ? QLatin1String("String") : QLatin1String("regexpr")).toString());
+        const QString string(xml.attributes().value((xml.name() == QLatin1String("RegExpr")) ? QLatin1String("String") : QLatin1String("regexpr")).toString());
 
         // validate regexp
-        const QRegularExpression regexp (string);
+        const QRegularExpression regexp(string);
         if (!regexp.isValid()) {
             qWarning() << hlFilename << "line" << xml.lineNumber() << "broken regex:" << string << "problem:" << regexp.errorString() << "at offset" << regexp.patternErrorOffset();
             return false;
@@ -163,6 +162,7 @@ bool checkSingleChars(const QString &hlFilename, QXmlStreamReader &xml)
         const QString c = xml.attributes().value(QLatin1String("char")).toString();
         if (c.size() != 1) {
             qWarning() << hlFilename << "line" << xml.lineNumber() << "'char' must contain exactly one char:" << c;
+            return false;
         }
     }
 
@@ -170,6 +170,7 @@ bool checkSingleChars(const QString &hlFilename, QXmlStreamReader &xml)
         const QString c = xml.attributes().value(QLatin1String("char1")).toString();
         if (c.size() != 1) {
             qWarning() << hlFilename << "line" << xml.lineNumber() << "'char1' must contain exactly one char:" << c;
+            return false;
         }
     }
 
@@ -201,20 +202,19 @@ class KeywordIncludeChecker
 public:
     void processElement(const QString &hlFilename, const QString &hlName, QXmlStreamReader &xml)
     {
-         if (xml.name() == QLatin1String("list")) {
-             auto &keywords = m_keywordMap[hlName];
-             keywords.filename = hlFilename;
-             auto name = xml.attributes().value(QLatin1String("name")).toString();
-             m_currentIncludes = &keywords.includes[name];
-         }
-         else if (xml.name() == QLatin1String("include")) {
-             if (!m_currentIncludes) {
-                 qWarning() << hlFilename << "line" << xml.lineNumber() << "<include> tag ouside <list>";
-                 m_success = false;
-             } else {
-                 m_currentIncludes->push_back({xml.lineNumber(), xml.readElementText()});
-             }
-         }
+        if (xml.name() == QLatin1String("list")) {
+            auto &keywords = m_keywordMap[hlName];
+            keywords.filename = hlFilename;
+            auto name = xml.attributes().value(QLatin1String("name")).toString();
+            m_currentIncludes = &keywords.includes[name];
+        } else if (xml.name() == QLatin1String("include")) {
+            if (!m_currentIncludes) {
+                qWarning() << hlFilename << "line" << xml.lineNumber() << "<include> tag ouside <list>";
+                m_success = false;
+            } else {
+                m_currentIncludes->push_back({xml.lineNumber(), xml.readElementText()});
+            }
+        }
     }
 
     bool check() const
@@ -230,8 +230,7 @@ public:
                     if (idx == -1) {
                         auto &keywordName = includes.key();
                         containsKeywordName = keywords.includes.contains(keywordName);
-                    }
-                    else {
+                    } else {
                         auto defName = include.name.mid(idx + 2);
                         auto listName = include.name.left(idx);
                         auto it = m_keywordMap.find(defName);
@@ -254,11 +253,9 @@ public:
     }
 
 private:
-    struct Keywords
-    {
+    struct Keywords {
         QString filename;
-        struct Include
-        {
+        struct Include {
             qint64 line;
             QString name;
         };
@@ -277,7 +274,8 @@ class KeywordChecker
 public:
     KeywordChecker(const QString &filename)
         : m_filename(filename)
-    {}
+    {
+    }
 
     void processElement(QXmlStreamReader &xml)
     {
@@ -285,6 +283,7 @@ public:
             const QString name = xml.attributes().value(QLatin1String("name")).toString();
             if (m_existingNames.contains(name)) {
                 qWarning() << m_filename << "list duplicate:" << name;
+                m_success = false;
             }
             m_existingNames.insert(name);
         } else if (xml.name() == QLatin1String("keyword")) {
@@ -296,7 +295,7 @@ public:
 
     bool check() const
     {
-        bool success = true;
+        bool success = m_success;
         const auto invalidNames = m_usedNames - m_existingNames;
         if (!invalidNames.isEmpty()) {
             qWarning() << m_filename << "Reference of non-existing keyword list:" << invalidNames;
@@ -306,6 +305,7 @@ public:
         const auto unusedNames = m_existingNames - m_usedNames;
         if (!unusedNames.isEmpty()) {
             qWarning() << m_filename << "Unused keyword lists:" << unusedNames;
+            success = false;
         }
 
         return success;
@@ -315,18 +315,31 @@ private:
     QString m_filename;
     QSet<QString> m_usedNames;
     QSet<QString> m_existingNames;
+    bool m_success = true;
 };
 
 /**
- * Helper class to search for non-existing contexts
+ * Helper class to search for non-existing contexts and invalid version
  */
 class ContextChecker
 {
 public:
+    void setKateVersion(const QStringRef &verStr, const QString &hlFilename, const QString &hlName)
+    {
+        const auto idx = verStr.indexOf(QLatin1Char('.'));
+        if (idx <= 0) {
+            qWarning() << hlFilename << "invalid kateversion" << verStr;
+            m_success = false;
+        } else {
+            auto &language = m_contextMap[hlName];
+            language.version = {verStr.left(idx).toInt(), verStr.mid(idx + 1).toInt()};
+        }
+    }
+
     void processElement(const QString &hlFilename, const QString &hlName, QXmlStreamReader &xml)
     {
         if (xml.name() == QLatin1String("context")) {
-            auto & language = m_contextMap[hlName];
+            auto &language = m_contextMap[hlName];
             language.hlFilename = hlFilename;
             const QString name = xml.attributes().value(QLatin1String("name")).toString();
             if (language.isFirstContext) {
@@ -336,22 +349,28 @@ public:
 
             if (language.existingContextNames.contains(name)) {
                 qWarning() << hlFilename << "Duplicate context:" << name;
+                m_success = false;
             } else {
                 language.existingContextNames.insert(name);
             }
 
             if (xml.attributes().value(QLatin1String("fallthroughContext")).toString() == QLatin1String("#stay")) {
                 qWarning() << hlFilename << "possible infinite loop due to fallthroughContext=\"#stay\" in context " << name;
+                m_success = false;
             }
 
             processContext(hlName, xml.attributes().value(QLatin1String("lineEndContext")).toString());
             processContext(hlName, xml.attributes().value(QLatin1String("lineEmptyContext")).toString());
             processContext(hlName, xml.attributes().value(QLatin1String("fallthroughContext")).toString());
+        } else if (xml.name() == QLatin1String("include")) {
+            // <include> tag inside <list>
+            processVersion(hlFilename, hlName, xml, {5, 53}, QLatin1String("<include>"));
         } else {
             if (xml.attributes().hasAttribute(QLatin1String("context"))) {
                 const QString context = xml.attributes().value(QLatin1String("context")).toString();
                 if (context.isEmpty()) {
                     qWarning() << hlFilename << "Missing context name in line" << xml.lineNumber();
+                    m_success = false;
                 } else {
                     processContext(hlName, context);
                 }
@@ -361,7 +380,29 @@ public:
 
     bool check() const
     {
-        bool success = true;
+        bool success = m_success;
+
+        // recursive search for the required miximal version
+        struct GetRequiredVersion {
+            QHash<const Language *, Version> versionMap;
+
+            Version operator()(const QHash<QString, Language> &contextMap, const Language &language)
+            {
+                auto &version = versionMap[&language];
+                if (version < language.version) {
+                    version = language.version;
+                    for (auto &languageName : language.usedLanguageName) {
+                        auto it = contextMap.find(languageName);
+                        if (it != contextMap.end()) {
+                            version = std::max(operator()(contextMap, *it), version);
+                        }
+                    }
+                }
+                return version;
+            };
+        };
+        GetRequiredVersion getRequiredVersion;
+
         for (auto &language : m_contextMap) {
             const auto invalidContextNames = language.usedContextNames - language.existingContextNames;
             if (!invalidContextNames.isEmpty()) {
@@ -374,6 +415,12 @@ public:
                 qWarning() << language.hlFilename << "Unused contexts:" << unusedNames;
                 success = false;
             }
+
+            auto requiredVersion = getRequiredVersion(m_contextMap, language);
+            if (language.version < requiredVersion) {
+                qWarning().nospace() << language.hlFilename << " depends on a language in version " << requiredVersion.majorRevision << "." << requiredVersion.minorRevision << ". Please, increase kateversion.";
+                success = false;
+            }
         }
 
         return success;
@@ -405,6 +452,7 @@ private:
             } else if (list.size() == 2) {
                 // specific context of other language, e.g. Comment##ISO C++
                 m_contextMap[list[1]].usedContextNames.insert(list[0]);
+                m_contextMap[language].usedLanguageName.insert(list[1]);
             }
             return;
         }
@@ -418,6 +466,35 @@ private:
     }
 
 private:
+    struct Version {
+        int majorRevision;
+        int minorRevision;
+
+        Version(int majorRevision = 0, int minorRevision = 0)
+            : majorRevision(majorRevision)
+            , minorRevision(minorRevision)
+        {
+        }
+
+        bool operator<(const Version &version) const
+        {
+            return majorRevision < version.majorRevision || (majorRevision == version.majorRevision && minorRevision < version.minorRevision);
+        }
+    };
+
+    void processVersion(const QString &hlFilename, const QString &hlName, QXmlStreamReader &xml, Version const &requiredVersion, QLatin1String item)
+    {
+        auto &language = m_contextMap[hlName];
+
+        if (language.version < requiredVersion) {
+            qWarning().nospace() << hlFilename << " " << item << " in line " << xml.lineNumber() << " is only available since version " << requiredVersion.majorRevision << "." << requiredVersion.minorRevision
+                                 << ". Please, increase kateversion.";
+            // update the version to cancel future warnings
+            language.version = requiredVersion;
+            m_success = false;
+        }
+    }
+
     class Language
     {
     public:
@@ -435,6 +512,12 @@ private:
 
         // holds all existing context names
         QSet<QString> existingContextNames;
+
+        // holds all existing language names
+        QSet<QString> usedLanguageName;
+
+        // kateversion language attribute
+        Version version;
     };
 
     /**
@@ -442,6 +525,7 @@ private:
      * Example key: "Doxygen"
      */
     QHash<QString, Language> m_contextMap;
+    bool m_success = true;
 };
 
 /**
@@ -452,7 +536,8 @@ class AttributeChecker
 public:
     AttributeChecker(const QString &filename)
         : m_filename(filename)
-    {}
+    {
+    }
 
     void processElement(QXmlStreamReader &xml)
     {
@@ -461,6 +546,7 @@ public:
             if (!name.isEmpty()) {
                 if (m_existingAttributeNames.contains(name)) {
                     qWarning() << m_filename << "itemData duplicate:" << name;
+                    m_success = false;
                 } else {
                     m_existingAttributeNames.insert(name);
                 }
@@ -469,6 +555,7 @@ public:
             const QString name = xml.attributes().value(QLatin1String("attribute")).toString();
             if (name.isEmpty()) {
                 qWarning() << m_filename << "specified attribute is empty:" << xml.name();
+                m_success = false;
             } else {
                 m_usedAttributeNames.insert(name);
             }
@@ -477,7 +564,7 @@ public:
 
     bool check() const
     {
-        bool success = true;
+        bool success = m_success;
         const auto invalidNames = m_usedAttributeNames - m_existingAttributeNames;
         if (!invalidNames.isEmpty()) {
             qWarning() << m_filename << "Reference of non-existing itemData attributes:" << invalidNames;
@@ -487,6 +574,7 @@ public:
         auto unusedNames = m_existingAttributeNames - m_usedAttributeNames;
         if (!unusedNames.isEmpty()) {
             qWarning() << m_filename << "Unused itemData:" << unusedNames;
+            success = false;
         }
 
         return success;
@@ -496,6 +584,7 @@ private:
     QString m_filename;
     QSet<QString> m_usedAttributeNames;
     QSet<QString> m_existingAttributeNames;
+    bool m_success = true;
 };
 
 }
@@ -528,9 +617,8 @@ int main(int argc, char *argv[])
     }
 
     // text attributes
-    const QStringList textAttributes = QStringList() << QStringLiteral("name") << QStringLiteral("section") << QStringLiteral("mimetype")
-            << QStringLiteral("extensions") << QStringLiteral("style")
-            << QStringLiteral("author") << QStringLiteral("license") << QStringLiteral("indenter");
+    const QStringList textAttributes = QStringList() << QStringLiteral("name") << QStringLiteral("section") << QStringLiteral("mimetype") << QStringLiteral("extensions") << QStringLiteral("style") << QStringLiteral("author")
+                                                     << QStringLiteral("license") << QStringLiteral("indenter");
 
     // index all given highlightings
     ContextChecker contextChecker;
@@ -540,7 +628,7 @@ int main(int argc, char *argv[])
     for (const QString &hlFilename : qAsConst(hlFilenames)) {
         QFile hlFile(hlFilename);
         if (!hlFile.open(QIODevice::ReadOnly)) {
-            qWarning ("Failed to open %s", qPrintable(hlFilename));
+            qWarning("Failed to open %s", qPrintable(hlFilename));
             anyError = 3;
             continue;
         }
@@ -594,8 +682,11 @@ int main(int argc, char *argv[])
 
         AttributeChecker attributeChecker(hlFilename);
         KeywordChecker keywordChecker(hlFilename);
+
         const QString hlName = hl[QStringLiteral("name")].toString();
 
+        contextChecker.setKateVersion(xml.attributes().value(QStringLiteral("kateversion")), hlFilename, hlName);
+
         // scan for broken regex or keywords with spaces
         while (!xml.atEnd()) {
             xml.readNext();
@@ -655,7 +746,6 @@ int main(int argc, char *argv[])
     if (!keywordIncludeChecker.check())
         anyError = 7;
 
-
     // bail out if any problem was seen
     if (anyError)
         return anyError;
